{-# OPTIONS -Wall #-}
module Language.Barracuda.Compiler
  ( module Language.Barracuda.Function
  , module Language.Barracuda.Compiler.Infrastructure
  , module Language.Barracuda.Compiler.Cpp
  , module Language.Barracuda.Compiler.Cuda
  , Doc
  , render
  , compile
  ) where

import Language.Barracuda.Compiler.Infrastructure
import Language.Barracuda.Function (Function)
import Language.Barracuda.Compiler.Cpp (compileCpp)
import Language.Barracuda.Compiler.Cuda (compileCuda)

import Data.Char (isAlphaNum, toUpper)
import System.FilePath (takeFileName)
import Text.PrettyPrint

compile :: CompileConfig -> (FilePath, FilePath) -> [Function] -> IO ()
compile config names@(hName, cName) funs = do
  let (hf, cf) = case config of
                   CppTarget {}  -> compileCpp config funs
                   CudaTarget {} -> compileCuda config funs
  msgWriting hName
  writeHeader hName hf
  msgWriting cName
  writeCode names cf

msgWriting :: FilePath -> IO ()
msgWriting fn = putStr "writing " >> putStrLn fn

writeHeader :: FilePath -> Doc -> IO ()
writeHeader hName hf = writeFileLn hName (render hf')
  where hf' = wrapFileInfo hName (wrapGuard hName hf)

writeCode :: (FilePath, FilePath) -> Doc -> IO ()
writeCode (hName, cName) cf = writeFileLn cName (render cf')
  where
    cf' = wrapFileInfo cName (include $+$ cf)
    include = text ("#include \"" ++ takeFileName hName ++ "\"")

writeFileLn :: FilePath -> String -> IO ()
writeFileLn f cs = writeFile f (cs ++ "\n")

wrapFileInfo :: FilePath -> Doc -> Doc
wrapFileInfo f d = wrap b d a
  where
    b = comment ("begin " ++ f)
    a = comment ("end " ++ f)

comment :: String -> Doc
comment t = text ("// " ++ t)

wrapGuard :: FilePath -> Doc -> Doc
wrapGuard f d = wrap b d a
  where
    b = text ("#ifndef " ++ gName) $+$ text ("#define " ++ gName)
    a = text "#endif"
    gName = map (replace . toUpper) f
    replace c = if isAlphaNum c then c else '_'

wrap :: Doc -> Doc -> Doc -> Doc
wrap before body after = before $+$ text "" $+$ body $+$ text "" $+$ after
