{-# LANGUAGE GADTs #-}
module Language.Barracuda.Compiler.ArgTypes
  (
    gpuScalarType
  , kernelScalarType
  , cpuScalarType

  , gpuArrayType
  , kernelArrayType
  , cpuArrayType

  , gpuOutputArrayType
  , kernelOutputArrayType
  , cpuOutputArrayType

  , gpuMatrixType
  , kernelMatrixType
  , cpuMatrixType

  , gpuOutputMatrixType
  , kernelOutputMatrixType
  , cpuOutputMatrixType

  , gpuArgList
  , kernelArgList
  , cpuArgList
  ) where

import Language.Barracuda.AST
import Language.Barracuda.Compiler.CppPrettyPrint
import Language.Barracuda.Function
import Util.PrettyPrintHelpers

import Text.PrettyPrint (Doc, text, (<+>), parens)

-- | Return the scalar type for a function argument with the given
-- witness type.
gpuScalarType :: ScalarWitness a -> Doc
gpuScalarType WB = tBool
gpuScalarType WF = tFloat
gpuScalarType WI = tInt

cpuScalarType :: ScalarWitness a -> Doc
cpuScalarType = gpuScalarType

kernelScalarType :: ScalarWitness a -> Doc
kernelScalarType = gpuScalarType

gpuArrayType :: ScalarWitness a -> Doc
gpuArrayType w =
  case w of
    WB -> text "gpu_bool_array"
    WF -> text "gpu_float_array"
    WI -> text "gpu_int_array"

gpuOutputArrayType :: ScalarWitness a -> Doc
gpuOutputArrayType w =
  case w of
    WB -> text "gpu_bool_array_view"
    WF -> text "gpu_float_array_view"
    WI -> text "gpu_int_array_view"

kernelArrayType :: ScalarWitness a -> Doc
kernelArrayType w =
  case w of
    WB -> text "kernel_bool_array"
    WF -> text "kernel_float_array"
    WI -> text "kernel_int_array"

kernelOutputArrayType :: ScalarWitness a -> Doc
kernelOutputArrayType w =
  case w of
    WB -> text "kernel_bool_array_view"
    WF -> text "kernel_float_array_view"
    WI -> text "kernel_int_array_view"

cpuArrayType :: ScalarWitness a -> Doc
cpuArrayType w =
  case w of
    WB -> text "bool_array"
    WF -> text "float_array"
    WI -> text "int_array"

cpuOutputArrayType :: ScalarWitness a -> Doc
cpuOutputArrayType w =
  case w of
    WB -> text "bool_array_view"
    WF -> text "float_array_view"
    WI -> text "int_array_view"

gpuMatrixType :: ScalarWitness a -> Doc
gpuMatrixType w =
  case w of
    WB -> text "gpu_bool_matrix"
    WF -> text "gpu_float_matrix"
    WI -> text "gpu_int_matrix"

gpuOutputMatrixType :: ScalarWitness a -> Doc
gpuOutputMatrixType w =
  case w of
    WB -> text "gpu_bool_matrix_view"
    WF -> text "gpu_float_matrix_view"
    WI -> text "gpu_int_matrix_view"

kernelMatrixType :: ScalarWitness a -> Doc
kernelMatrixType w =
  case w of
    WB -> text "kernel_bool_matrix"
    WF -> text "kernel_float_matrix"
    WI -> text "kernel_int_matrix"

kernelOutputMatrixType :: ScalarWitness a -> Doc
kernelOutputMatrixType w =
  case w of
    WB -> text "kernel_bool_matrix_view"
    WF -> text "kernel_float_matrix_view"
    WI -> text "kernel_int_matrix_view"

cpuMatrixType :: ScalarWitness a -> Doc
cpuMatrixType w =
  case w of
    WB -> text "bool_matrix"
    WF -> text "float_matrix"
    WI -> text "int_matrix"

cpuOutputMatrixType :: ScalarWitness a -> Doc
cpuOutputMatrixType w =
  case w of
    WB -> text "bool_matrix_view"
    WF -> text "float_matrix_view"
    WI -> text "int_matrix_view"

gpuArgList :: [Arg] -> Arg -> Doc
gpuArgList is o = parens $ commaSep (map inputArg is ++ [outputArg o])
  where
    inputArg :: Arg -> Doc
    inputArg (SArg sw n) = qConst <+> gpuScalarType sw <+> var n
    inputArg (AArg sw n) = qConst <+> gpuArrayType sw <+> refVar n
    inputArg (MArg sw n) = qConst <+> gpuMatrixType sw <+> refVar n
    
    outputArg :: Arg -> Doc
    outputArg (SArg sw n) = gpuScalarType sw <+> refVar n
    outputArg (AArg sw n) = gpuOutputArrayType sw <+> var n
    outputArg (MArg sw n) = gpuOutputMatrixType sw <+> var n

kernelArgList :: [Arg] -> Arg -> Doc
kernelArgList is o = parens $ commaSep (map inputArg is ++ [outputArg o])
  where
    inputArg :: Arg -> Doc
    inputArg (SArg sw n) = qConst <+> kernelScalarType sw <+> var n
    inputArg (AArg sw n) = qConst <+> kernelArrayType sw <+> var n
    inputArg (MArg sw n) = qConst <+> kernelMatrixType sw <+> var n
    
    outputArg :: Arg -> Doc
    outputArg (SArg {})   = error "kernelArgList: scalar kernel output?"
    outputArg (AArg sw n) = kernelOutputArrayType sw <+> var n
    outputArg (MArg sw n) = kernelOutputMatrixType sw <+> var n

cpuArgList :: [Arg] -> Arg -> Doc
cpuArgList is o = parens $ commaSep (map inputArg is ++ [outputArg o])
  where
    inputArg :: Arg -> Doc
    inputArg (SArg sw n) = qConst <+> cpuScalarType sw <+> var n
    inputArg (AArg sw n) = qConst <+> cpuArrayType sw <+> refVar n
    inputArg (MArg sw n) = qConst <+> cpuMatrixType sw <+> refVar n
    
    outputArg :: Arg -> Doc
    outputArg (SArg sw n) = cpuScalarType sw <+> refVar n
    outputArg (AArg sw n) = cpuOutputArrayType sw <+> var n
    outputArg (MArg sw n) = cpuOutputMatrixType sw <+> var n
