{-# LANGUAGE GADTs, ScopedTypeVariables, Rank2Types #-}
{-# OPTIONS -Wall #-}
module Language.Barracuda.Compiler.Cuda ( compileCuda ) where

------------------------------------------------------------------------
-- Imports
------------------------------------------------------------------------
import Language.Barracuda.AST
import Language.Barracuda.SmartConstructors hiding (int, bool, slet)
import qualified Language.Barracuda.SmartConstructors as Smart
import Language.Barracuda.Function
import Language.Barracuda.Compiler.Aliasing
import Language.Barracuda.Compiler.ArgTypes
import Language.Barracuda.Compiler.FreeVariables
import Language.Barracuda.Compiler.Infrastructure
import Language.Barracuda.Compiler.CppPrettyPrint
import Util.PrettyPrintHelpers

import Control.Monad (liftM, liftM3)
import Control.Monad.Writer (WriterT, runWriterT, tell, lift)
import qualified Data.Map as Map
import Data.Monoid (Monoid, mappend, mempty, mconcat)
import Text.PrettyPrint
-- import Debug.Trace (trace)

------------------------------------------------------------------------
-- Types & Instances
------------------------------------------------------------------------
newtype Statement = Statement { unStatement :: Doc }

instance Monoid Statement where
  mempty = Statement empty
  mappend (Statement s1) (Statement s2) = Statement (s1 $+$ s2)

-- newtype Expr = Expr { unExpr :: Doc }
-- newtype Type = Type { unType :: Doc }
-- newtype Qualifier = Qualifier { unQualifier :: Doc }

------------------------------------------------------------------------
-- Target-specific combinators for generation of pretty-printed C++
------------------------------------------------------------------------
qShared :: Doc
qShared = text "__shared__"

qGlobal :: Doc
qGlobal = text "__global__"

function :: Id -> [Arg] -> Arg -> Doc -> Doc
function fn is o b = funSig fn is o $+$ block b

funSig :: Id -> [Arg] -> Arg -> Doc
funSig fn is o = tVoid <+> var fn <+> gpuArgList is o

------------------------------------------------------------------------
-- Compilation code
------------------------------------------------------------------------
compileCuda :: CompileConfig -> [Function] -> (Doc, Doc)
compileCuda config funs = (hfile, cfile)
  where
    hfile = vsep [header, declarations, footer]
    cfile = vsep [header, kernels, functions, footer]
    header = banner "header & includes" $+$ include "barracuda.hpp"
    footer = banner "footer"
    kernels = banner "kernels" $+$ vsep (map unKernel ks)
    functions = banner "functions" $+$ vsep fs
    declarations = banner "declarations" $+$ vsep ds
    (ds, fs, ks) = compileCudaFuns config funs

compileCudaFuns :: CompileConfig -> [Function] -> ([Doc], [Doc], [Kernel])
compileCudaFuns config funs = (ds, fs, concat kss)
  where
    (ds, fs, kss) = unzip3 funs'
    funs' = runCompile config (mapM compileFunction funs)

-- | Compiles a Barracuda function annotated with name information,
-- returning a function declaration, the compiled function, and a list
-- of compiled CUDA kernels.
compileFunction :: Function -> Compile (Doc, Doc, [Kernel])
compileFunction (Function f n is o) = do
  let (is', o', f') = instantiate (is ++ [o]) f
  (Statement f'', ks) <- compileSomeExpr o' f'
  return (funSig n is' o' <> semi, function n is' o' f'', ks)

-- | Compiles an assignment statement with the given LHS identifier
-- and RHS expression.  The compiled expression and a list of compiled
-- kernels is returned.
compileSomeExpr :: Arg -> SomeExpr -> Compile (Statement, [Kernel])
compileSomeExpr outArg expr =
  let outName = argName outArg
  in case expr of
       SExpr e -> compileAssignSExpr outName e
       AExpr e -> compileAssignAExpr outName e
       MExpr e -> compileAssignMExpr outName e

------------------------------------------------------------------------
-- Common code for scalar expression compilation
------------------------------------------------------------------------
data ScalarCompiler m = ScalarCompiler {
    areducer :: Reducer m AExpr
  , mreducer :: Reducer m MExpr
  , slet     :: forall a b. (Scalar a, Scalar b) =>
                SExpr a -> (SExpr a -> SExpr b) -> m Doc
  , aIdx :: forall a. (Scalar a) =>
            ScalarWitness a -> Id -> SExpr Int -> m Doc
  , mIdx :: forall a. (Scalar a) =>
            ScalarWitness a -> Id -> SExpr Int -> SExpr Int -> m Doc
  }

type Reducer m e = forall t. (Scalar t)
  => (SExpr t -> SExpr t -> SExpr t)
  -> SExpr t
  -> e t
  -> m Doc

compileScalar
  :: forall m a. Monad m
  => ScalarCompiler m
  -> SExpr a                       -- ^ expression to compile
  -> m Doc
compileScalar dict = go
  where
    go :: forall t. SExpr t -> m Doc
    go expr =
      case expr of
        SLet b f -> slet dict b f
        SVar n -> return (var n)
        Prim1 o e -> do
            e' <- go e
            return $ funcall (lookupPrimOp1Function o) [e']
        Prim2 o l r -> do
            l' <- go l
            r' <- go r
            let op = text (lookupPrimOp2Infix o)
            return $ sep [parenthesize l l', op, parenthesize r r']
        SCond p t f -> do
            p' <- liftM (parenthesize p) (go p)
            t' <- liftM (parenthesize t) (go t)
            f' <- liftM (parenthesize f) (go f)
            return $ ternary p' t' f'
        SConst c ->
          case c of
            SCF f -> return (litFloat f)
            SCI i -> return (int i)
            SCB b -> return (bool b)
        AReduce f i a -> areducer dict f i a
        AIdx n i -> aIdx dict (sexprType expr) n i
        ALen t ->
          case t of
            AVar n -> return $ method (var n) "length" []
            ASliceVar si _ -> go $ sliceLength si
            AMap _f a -> go (alength a)
            AMap2 _f a _b -> go (alength a)
            AMap3 _f a _b _c -> go (alength a)
        MReduce f i a -> mreducer dict f i a
        MIdx n i j -> mIdx dict (sexprType expr) n i j
        MRows e ->
          case e of
            MVar n -> return $ method (var n) "rows" []
            MSliceVar si _ -> go (msliceRows si)
            MMap _f a -> go (mrows a)
            MMap2 _f a _b -> go (mrows a)
        MCols e ->
          case e of
            MVar n -> return $ method (var n) "cols" []
            MSliceVar si _ -> go (msliceCols si)
            MMap _f a -> go (mcols a)
            MMap2 _f a _b -> go (mcols a)
          

-- | Generates code to loop from 0 to the given stop value, exclusive.
-- Used to loop over array expressions.  Returns a function for the
-- loop body and the name of the indexing variable.
loopFromZeroTo
  :: forall m. Monad m
  => (forall t. Compile t -> m t)
  -> (forall t. Scalar t => SExpr t -> m Doc)
  -> SExpr Int                       -- ^  the stop value
  -> m (Doc -> Doc, Id)
loopFromZeroTo l k stop = do
  idxName <- l (uniqueId "idx")
  stopVal <- k stop
  let idxVar = var idxName
      loop = forLoop (tInt <+> idxVar <+> equals <+> int 0)
                    (idxVar <+> char '<' <+> stopVal)
                    (idxVar <+> equals <+> idxVar <+> char '+' <+> int 1)
  return (loop, idxName)

------------------------------------------------------------------------
-- Top-Level Scalar Expression Compilation
------------------------------------------------------------------------
newtype Kernel = Kernel { unKernel :: Doc }

type CompileScalar a = WriterT (Statement, [Kernel]) Compile a

runCompileScalar :: CompileScalar a -> Compile (a, Statement, [Kernel])
runCompileScalar compile = do
  (result, (hoisted, kernels)) <- runWriterT compile
  return (result, hoisted, kernels)

emitKernel :: Kernel -> CompileScalar ()
emitKernel k = tell (mempty, [k])

hoistStatement :: Statement -> CompileScalar ()
hoistStatement s = tell (s, mempty)

liftCompile :: Compile a -> CompileScalar a
liftCompile = lift

-- | Compiles a top-level scalar assignment statement.  The compiled
-- expression and a list of compiled kernels is returned.
compileAssignSExpr
  :: (Scalar a)
  => Id             -- ^  the name of the variable to assign
  -> SExpr a        -- ^  the scalar expression to compile
  -> Compile (Statement, [Kernel])
compileAssignSExpr outName expr = do
  (expr', hoisted, kernels) <- runCompileScalar (compileSExpr expr)
  let assignment = Statement $ assign (var outName) expr'
  return (hoisted `mappend` assignment, kernels)

compileSExpr :: (Scalar a) => SExpr a -> CompileScalar Doc
compileSExpr = compileScalar dict
  where
    dict = ScalarCompiler compileAReduce compileMReduce sLet aidx midx
    sLet b f = do
      b' <- compileSExpr b
      bindName <- liftCompile (uniqueId "let")
      let bindDecl = qConst <+> cpuScalarType (sexprType b) <+>
                     assign (var bindName) b'
      hoistStatement (Statement bindDecl)
      compileSExpr (f $ SVar bindName)

    aidx _w n i = liftM (index (var n)) (compileSExpr i)

    midx _w n i j = do i' <- compileSExpr i
                       j' <- compileSExpr j
                       return $ funcall n [i', j']


compileTempSExpr :: (Scalar a) => String -> SExpr a -> CompileScalar Doc
compileTempSExpr prefix expr = do
  name <- liftCompile (uniqueId prefix)
  expr' <- compileSExpr expr
  let bind = qConst <+> gpuScalarType (sexprType expr)
                    <+> var name <> parens expr' <> semi
  hoistStatement (Statement bind)
  return (var name)

compileAReduce :: Reducer CompileScalar AExpr
compileAReduce fun iVal aexpr = do
  -- block dimensions
  bd <- compileTempSExpr "blockDim" (blockDims :: SExpr Int)

  let elemsPerThread = 16
  -- grid dimensions
  gd <- compileTempSExpr "gridDims" $
          minVal 64
                 (maxVal 1 (alength aexpr `idiv` blockDims `idiv` elemsPerThread))

  -- scratch array
  scratchName <- liftCompile (uniqueId "reduce_acc_scratch")
  let scratchDecl = gpuArrayType (aexprElemType aexpr) <+>
                    var scratchName <> parens gd <> semi
  hoistStatement (Statement scratchDecl)

  -- kernel
  (kernelName, kernel, kernelArgs, outArg) <- liftCompile $
                                                compileAReduceKernel
                                                  scratchName fun aexpr
  emitKernel kernel

  -- invocation
  let kparams = text "<<<" <> commaSep [gd, bd] <> text ">>>"
  let params = kernelActualParamList kernelArgs outArg
  let invocation = text kernelName <> kparams <> params <> semi
  hoistStatement (Statement invocation)

  -- reduce scratch array
  -- cpu scratch array
  cpuScratch <- liftCompile (uniqueId "reduce_acc_cpu_scratch")
  let cpuScratchDecl = cpuArrayType (aexprElemType aexpr) <+> var cpuScratch <>
                       parens (var scratchName) <> semi
  hoistStatement (Statement cpuScratchDecl)

  -- accumulator for reduction of scratch array
  iVal' <- compileSExpr iVal
  accName <- liftCompile (uniqueId "reduce_acc")
  let accDecl = gpuScalarType (sexprType iVal) <+> assign (var accName) iVal'
  hoistStatement (Statement accDecl)

  -- loop over scratch array, doing reduction on host
  let scratchExpr = AVar cpuScratch `asTypeOf` aexpr
  (loop, idxVarName) <- loopFromZeroTo
                          liftCompile
                          compileSExpr
                          (alength scratchExpr)
  let fInst = fun (SVar accName) (scratchExpr ! SVar idxVarName)
  -- TODO: if the hoisted portion is safe to pull out of the loop, do so.
  (f', hoisted, ks) <- liftCompile . runCompileScalar . compileSExpr $ fInst
  mapM_ emitKernel ks
  let body = unStatement hoisted $+$ assign (var accName) f'
  hoistStatement (Statement $ loop body)

  return (var accName)
  
compileAReduceKernel
  :: forall a. (Scalar a)
  => Id
  -> (SExpr a -> SExpr a -> SExpr a)
  -> AExpr a
  -> Compile (Id, Kernel, [Arg], Arg)
compileAReduceKernel accName fun aexpr = do
  aliasInfo <- aexprAliases aexpr
  let elemType = aexprElemType aexpr

  let tid = SVar "threadIdx.x" :: SExpr Int

  scratchName <- uniqueId "scratch"
  let scratchDecl = qShared <+>
                    gpuScalarType elemType <+>
                    var scratchName <> brackets (int blockDims) <> semi
  let scratch = AVar scratchName :: AExpr a

  gridSizeName <- uniqueId "grid_size"
  let gridSizeDecl = qConst <+> tInt <+> var gridSizeName <+> equals <+>
                       int blockDims <+> char '*' <+> int 2 <+> char '*' <+>
                       var "gridDim.x" <> semi
  -- let gridSize = SVar gridSizeName :: SExpr Int

  idxName <- uniqueId "i"
  let idxDecl = tInt <+>
                  assign (var idxName)
                         (var "blockIdx.x" <+>
                          char '*' <+>
                          int blockDims <+>
                          char '*' <+>
                          int 2 <+>
                          char '+' <+>
                          var "threadIdx.x")

  let idx = SVar idxName :: SExpr Int

  ----------------------------------------
  -- some utility functions
  ----------------------------------------
  let assignScratch rhs = do
        (rhs', hoisted) <- runWriterT (compileKernelSExpr aliasInfo rhs)
        let lhs = index (var scratchName) (var "threadIdx.x")
        return $ hoisted `mappend` Statement (assign lhs rhs')
  let cond p b = do
      (p', ph) <- runWriterT (compileKernelSExpr aliasInfo p)
      let cond' = Statement $ ifBlock p' $ unStatement b
      return (ph `mappend` cond')
  let condAssignScratch p rhs = cond p =<< assignScratch rhs

  let blockSize = blockDims :: SExpr Int
  let numElems = alength aexpr
  let incrIdx = Statement $
                  assign
                    (var idxName)
                    (var idxName <+> char '+' <+> var gridSizeName)
  let sync = Statement $ text "__syncthreads();"

  ----------------------------------------
  -- compile kernel body
  ----------------------------------------
  load1 <- condAssignScratch (idx <* numElems) (aexpr ! idx)
  load2 <- condAssignScratch
             (idx + blockSize <* numElems)
             (fun (scratch ! tid) (aexpr ! (idx + blockSize)))
  reduce1 <- do
    (test, testh) <- runWriterT (compileKernelSExpr aliasInfo
                                                        (idx <* numElems))
    assign1 <- assignScratch $ fun (scratch ! tid) (aexpr ! idx)
    assign2 <- condAssignScratch
                 (idx + blockSize <* numElems)
                 (fun (scratch ! tid) (aexpr ! (idx + blockSize)))
    let loop = Statement $
                 whileLoop
                   test
                   (unStatement $ mconcat [assign1, assign2, incrIdx])
    return $ testh `mappend` loop
  reduce2 <- do
    let step size = do
          assignment <- let size' = Smart.int size
                        in condAssignScratch
                             (tid <* size' &&* tid + size' <* numElems)
                             (fun (scratch ! tid) (scratch ! (tid + size')))
          return $ Statement $ ifBlock
                     (int blockDims <+> text ">=" <+> int (2 * size))
                     (unStatement $ assignment `mappend` sync)
    liftM mconcat $ mapM step [256, 128, 64]
  reduce3 <- do
    let step size =
          let size' = Smart.int size
          in condAssignScratch
             (blockSize >=* size' * 2 &&* tid + size' <* numElems)
             (fun (scratch ! tid) (scratch ! (tid + size')))
    body <- liftM mconcat $ mapM step [32, 16, 8, 4, 2, 1]
    return $ Statement $
               ifBlock (var "threadIdx.x" <+> text "<" <+> int 32)
               (unStatement body)
  let output = Statement $ ifBlock
                 (var "threadIdx.x" <+> text "==" <+> int 0)
                 (assign (index (var accName) (var "blockIdx.x"))
                         (index (var scratchName) (int 0)))
  shmemLoad <- compileLoadSharedMemoryArray aliasInfo
  let kernelBody = mconcat [shmemLoad, load1, load2, incrIdx,
                            reduce1, sync, reduce2, reduce3,
                            output
                            ]

  ----------------------------------------
  -- assemble values to return
  ----------------------------------------
  kernelName <- uniqueId "kernel"
  freeVars <- getNonCudaFreeVars [scratchName, idxName]
              -- some semi-arbitrary instantiation
              (fun (scratch ! tid) (aexpr ! idx))
  let outputArg = AArg elemType accName
  let decls = mconcat . map Statement $ [scratchDecl, gridSizeDecl, idxDecl]
  let kernel = Kernel $
                 qStatic <+> qGlobal <+> tVoid $+$
                 var kernelName <+>
                 kernelArgList freeVars outputArg $+$
                 block (unStatement $ decls `mappend` kernelBody)
  return (kernelName, kernel, freeVars, outputArg)


compileMReduce :: Reducer CompileScalar MExpr
compileMReduce fun iVal mexpr = do
  -- grid dimensions
  let gridCols = (mcols mexpr + (blockDimsX - 1)) `idiv` blockDimsX
  let gridRows = (mrows mexpr + (blockDimsY - 1)) `idiv` blockDimsY
  gdX <- compileSExpr gridCols
  gdY <- compileSExpr gridRows
  gridDimsName <- liftCompile (uniqueId "gridDims")
  hoistStatement $ Statement $
                 qConst <+> text "dim3" <+>
                 dim3 gridDimsName (gdX, gdY, int 1) <> semi

  -- block dimensions
  blockDimsName <- liftCompile (uniqueId "blockDims")
  hoistStatement $ Statement $
                 qConst <+> text "dim3" <+>
                 dim3 blockDimsName (int blockDimsX, int blockDimsY, int 1) <>
                 semi

  -- scratch array
  scratchName <- liftCompile (uniqueId "reduce_acc_scratch")
  let scratchDecl = gpuMatrixType (mexprElemType mexpr) <+>
                    var scratchName <> parens (commaSep [gdY, gdX]) <> semi
  hoistStatement (Statement scratchDecl)

  -- kernel
  (kernelName, kernel, kInputs, kOutput) <- liftCompile $
                                                compileMReduceKernel
                                                  scratchName fun mexpr
  emitKernel kernel

  -- invocation
  let kparams = text "<<<" <>
                commaSep [var gridDimsName, var blockDimsName] <>
                text ">>>"
  let params = kernelActualParamList kInputs kOutput
  let invocation = text kernelName <> kparams <> params <> semi
  hoistStatement (Statement invocation)

  -- reduce scratch array
  -- cpu scratch array
  cpuScratch <- liftCompile (uniqueId "reduce_acc_cpu_scratch")
  let cpuScratchDecl = cpuMatrixType (mexprElemType mexpr) <+> var cpuScratch <>
                       parens (var scratchName) <> semi
  hoistStatement (Statement cpuScratchDecl)

  -- accumulator for reduction of scratch array
  iVal' <- compileSExpr iVal
  accName <- liftCompile (uniqueId "reduce_acc")
  let accDecl = gpuScalarType (sexprType iVal) <+> assign (var accName) iVal'
  hoistStatement (Statement accDecl)

  -- loop over scratch array, doing reduction on host
  let scratchExpr = MVar cpuScratch `asTypeOf` mexpr
  (loopOuter, rowIdxName) <- loopFromZeroTo
                             liftCompile
                             compileSExpr
                             (mrows scratchExpr)
  (loopInner, colIdxName) <- loopFromZeroTo
                             liftCompile
                             compileSExpr
                             (mcols scratchExpr)
  let fInst = let idx = (SVar rowIdxName, SVar colIdxName)
              in fun (SVar accName) (scratchExpr ! idx)
  -- TODO: if the hoisted portion is safe to pull out of the loop, do so.
  (f', hoisted, ks) <- liftCompile . runCompileScalar . compileSExpr $ fInst
  mapM_ emitKernel ks
  let body = unStatement hoisted $+$ assign (var accName) f'
  hoistStatement (Statement $ loopOuter $ loopInner body)

  return (var accName)


compileMReduceKernel
  :: forall a. (Scalar a)
  => Id
  -> (SExpr a -> SExpr a -> SExpr a)
  -> MExpr a
  -> Compile (Id, Kernel, [Arg], Arg)
compileMReduceKernel accName fun mexpr = do
  aliasInfo <- mexprAliases mexpr

  let elemType = mexprElemType mexpr
  let tidX = SVar "threadIdx.x" :: SExpr Int
  let tidY = SVar "threadIdx.y" :: SExpr Int

  scratchName <- uniqueId "scratch"
  let scratchDecl = Statement $
                    qShared <+>
                    gpuScalarType elemType <+>
                    var scratchName <>
                    brackets (int blockDimsY) <>
                    brackets (int blockDimsX) <> semi
  let scratch = MVar scratchName :: MExpr a

  let blockIdxY = SVar "blockIdx.y" :: SExpr Int
  let blockIdxX = SVar "blockIdx.x" :: SExpr Int

  (yRhs, yRhsH) <- runWriterT $ compileKernelSExpr aliasInfo
                                         (blockIdxY * blockDimsY + tidY)
  yName <- uniqueId "y"
  let yDecl = yRhsH `mappend`
              Statement (qConst <+> tInt <+> assign (var yName) yRhs)
  let yVar = SVar yName :: SExpr Int

  (xRhs, xRhsH) <- runWriterT $ compileKernelSExpr aliasInfo
                                         (blockIdxX * blockDimsX + tidX)
  xName <- uniqueId "x"
  let xDecl = xRhsH `mappend`
              Statement (qConst <+> tInt <+> assign (var xName) xRhs)
  let xVar = SVar xName :: SExpr Int

  let idx = (yVar, xVar)

  (validRhs, validRhsH) <- runWriterT $ compileKernelSExpr aliasInfo
                                         (yVar <* mrows mexpr &&*
                                          xVar <* mcols mexpr)
  validName <- uniqueId "valid_tid"
  let validDecl = validRhsH `mappend`
                  Statement (qConst <+> tBool <+>
                             assign (var validName) validRhs)
  let validVar = SVar validName :: SExpr Bool

  ----------------------------------------
  -- some utility functions
  ----------------------------------------
  let assignScratch rhs = do
        (rhs', hoisted) <- runWriterT (compileKernelSExpr aliasInfo rhs)
        let lhs = index
                  (index (var scratchName) (var "threadIdx.y"))
                  (var "threadIdx.x")
        return $ hoisted `mappend` Statement (assign lhs rhs')
  let if' p b = do
      (p', ph) <- runWriterT (compileKernelSExpr aliasInfo p)
      let cond' = Statement $ ifBlock p' $ unStatement b
      return (ph `mappend` cond')
  let condAssignScratch p rhs = if' p =<< assignScratch rhs
  let syncThreads = Statement $ text "__syncthreads();"

  ----------------------------------------
  -- body code
  ----------------------------------------
  load1 <- condAssignScratch validVar (mexpr ! idx)
  reduceRows <- do
      idxName <- uniqueId "i"
      let idxVar = SVar idxName :: SExpr Int
      (test, testH) <- runWriterT $ compileKernelSExpr aliasInfo $
                       (validVar &&* tidY <* idxVar &&*
                        yVar + idxVar <* mrows mexpr)
      let scratchAtDoc x y = var scratchName <> brackets y <> brackets x
      let scratchAt x y = SVar (render $ scratchAtDoc x y) :: SExpr a
      (rhs, rhsH) <- let tx = var "threadIdx.x"
                         ty = var "threadIdx.y"
                         ty' = ty <+> char '+' <+> var idxName
                     in runWriterT $ compileKernelSExpr aliasInfo $
                        fun (scratchAt tx ty) (scratchAt tx ty')
      let cond = Statement $ ifBlock test $ assign
                 (scratchAtDoc (var "threadIdx.x") (var "threadIdx.y"))
                 rhs
      let loop = forLoop (tInt <+> var idxName <+>
                          equals <+>
                          int blockDimsY <+> char '/' <+> int 2)
                         (var idxName <+> char '>' <+> int 0)
                         (var idxName <+> text "/=" <+> int 2)
                         (unStatement $
                            mconcat [testH, rhsH, cond, syncThreads])
      return $ Statement loop
  reduceCols <- do
      idxName <- uniqueId "i"
      let idxVar = SVar idxName :: SExpr Int
      (test, testH) <- runWriterT $ compileKernelSExpr aliasInfo $
                       (validVar &&* tidY ==* 0 &&*
                        tidX <* idxVar &&* xVar + idxVar <* mcols mexpr)
      let scratchAtDoc x y = var scratchName <> brackets y <> brackets x
      let scratchAt x y = SVar (render $ scratchAtDoc x y) :: SExpr a
      (rhs, rhsH) <- let tx = var "threadIdx.x"
                         ty = int 0
                         tx' = tx <+> char '+' <+> var idxName
                     in runWriterT $ compileKernelSExpr aliasInfo $
                        fun (scratchAt tx ty) (scratchAt tx' ty)
      let cond = Statement $ ifBlock test $ assign
                 (scratchAtDoc (var "threadIdx.x") (int 0))
                 rhs
      let loop = forLoop (tInt <+> var idxName <+>
                          equals <+>
                          int blockDimsX <+> char '/' <+> int 2)
                         (var idxName <+> char '>' <+> int 0)
                         (var idxName <+> text "/=" <+> int 2)
                         (unStatement $
                            mconcat [testH, rhsH, cond, syncThreads])
      return $ Statement loop
  output <- do
      let accVar = MVar accName :: MExpr a
      let lhs = accVar ! (SVar "blockIdx.y", SVar "blockIdx.x")
      (lhs', lhsH) <- runWriterT $ compileKernelSExpr aliasInfo lhs
      let test = var "threadIdx.x" <+> text "==" <+> int 0 <+> text "&&" <+>
                 var "threadIdx.y" <+> text "==" <+> int 0
      let rhs = var scratchName <> brackets (int 0) <> brackets (int 0)
      return . mappend lhsH . Statement . ifBlock test $ (assign lhs' rhs)
  shmemLoad <- compileLoadSharedMemoryMatrix aliasInfo
  let kernelBody = mconcat [shmemLoad, load1, syncThreads, reduceRows,
                            reduceCols, output]

  ----------------------------------------
  -- assemble values to return
  ----------------------------------------
  kernelName <- uniqueId "kernel"
  freeVars <- getNonCudaFreeVars [scratchName, xName, yName, validName]
              -- some semi-arbitrary instantiation
              (fun (scratch ! (tidY, tidX)) (mexpr ! idx))
  let outputArg = MArg elemType accName
  let decls = mconcat [scratchDecl, yDecl, xDecl, validDecl]
  let kernel = Kernel $
                 qStatic <+> qGlobal <+> tVoid $+$
                 var kernelName <+>
                 kernelArgList freeVars outputArg $+$
                 block (unStatement $ decls `mappend` kernelBody)
  return (kernelName, kernel, freeVars, outputArg)

  


------------------------------------------------------------------------
-- Kernel Scalar Expression Compilation
------------------------------------------------------------------------
type CompileKernelScalar a = WriterT Statement Compile a

compileKernelSExpr
  :: (Scalar a)
  => AliasInfo
  -> SExpr a
  -> CompileKernelScalar Doc
compileKernelSExpr aliasInfo sexpr = do
  shmemOpt <- lift useSharedMemory
  let dict = if shmemOpt
             then ScalarCompiler aReduce mReduce sLet aidxShmem midxShmem
             else ScalarCompiler aReduce mReduce sLet aidxNoshmem midxNoshmem
  compileScalar dict sexpr

  where
    loopZeroTo = loopFromZeroTo lift (compileKernelSExpr aliasInfo)

    aReduce f i aexpr = do
      accVarName <- lift (uniqueId "reduce_acc")
      i' <- compileKernelSExpr aliasInfo i
      (loop, idxVarName) <- loopZeroTo (alength aexpr)
      let fInst = f (SVar accVarName) (aexpr ! SVar idxVarName)
      f' <- compileKernelSExpr aliasInfo fInst
      let accVar     = var accVarName
          accVarDecl = gpuScalarType (sexprType i) <+> assign accVar i'
      let body = assign accVar f'
      tell (Statement accVarDecl)
      tell (Statement $ ndpWarning $+$ loop body)
      return (var accVarName)

    mReduce f i mexpr = do
      accVarName <- lift (uniqueId "reduce_acc")
      i' <- compileKernelSExpr aliasInfo i
      (loopOuter, rowIdxName) <- loopZeroTo (mrows mexpr)
      (loopInner, colIdxName) <- loopZeroTo (mcols mexpr)
      let mVal = mexpr ! (SVar rowIdxName, SVar colIdxName)
      let fInst = f (SVar accVarName) mVal
      f' <- compileKernelSExpr aliasInfo fInst
      let accVar = var accVarName
      let accVarDecl = gpuScalarType (sexprType i) <+> assign accVar i'
      let body = assign accVar f'
      tell (Statement accVarDecl)
      tell (Statement (ndpWarning $+$ loopOuter (loopInner $ body)))
      return accVar

    ndpWarning = text "#warning" <+> doubleQuotes
                 (text "Generated code performs a nested reduction," <+>
                  text "which is almost surely not what you want.")

    sLet b f = do
      b' <- compileKernelSExpr aliasInfo b
      bindName <- lift (uniqueId "let")
      let bindDecl = qConst <+> gpuScalarType (sexprType b) <+>
                     assign (var bindName) b'
      tell (Statement bindDecl)
      compileKernelSExpr aliasInfo (f $ SVar bindName)

    arrayUseShmem = arrayShmemTest aliasInfo
    matrixUseShmem = matrixShmemTest aliasInfo

    aidxShmem w n i
        | arrayUseShmem (n, SomeScalarWitness w) = do
              i' <- compileKernelSExpr aliasInfo i
              let p = text "blockIdx.x *" <+> int blockDims <+>
                      text "<=" <+> i' <+>
                      text "&&" <+> i' <+>
                      char '<' <+> int blockDims <+> text "* (blockIdx.x + 1)"
              let t = index (var (n ++ "_shared"))
                            (i' <+> text "- blockIdx.x *" <+> int blockDims)
              let f = index (var n) i'
              return . parens $ ternary p t f
        | otherwise = aidxNoshmem w n i

    aidxNoshmem _w n i = do
      i' <- compileKernelSExpr aliasInfo i
      return $ index (var n) i'

    midxShmem w n i j
        | matrixUseShmem (n, SomeScalarWitness w) = do
              i' <- compileKernelSExpr aliasInfo i
              j' <- compileKernelSExpr aliasInfo j
              let px = text "blockIdx.x *" <+> int blockDimsX <+>
                       text "<=" <+> j' <+>
                       text "&&" <+> j' <+>
                       char '<' <+> int blockDimsX <+> text "* (blockIdx.x + 1)"
              let py = text "blockIdx.y *" <+> int blockDimsY <+>
                       text "<=" <+> i' <+>
                       text "&&" <+> i' <+>
                       char '<' <+> int blockDimsY <+> text "* (blockIdx.y + 1)"
              let p = px <+> text "&&" <+> py
              let t = var (n ++ "_shared") <>
                      brackets (i' <+> text "- blockIdx.y *" <+> int blockDimsY) <>
                      brackets (j' <+> text "- blockIdx.x *" <+> int blockDimsX)
              let f = funcall n [i', j']
              return . parens $ ternary p t f
        | otherwise = midxNoshmem w n i j

    midxNoshmem _w n i j = do
      i' <- compileKernelSExpr aliasInfo i
      j' <- compileKernelSExpr aliasInfo j
      return $ funcall n [i', j']


------------------------------------------------------------------------
-- Top-Level Array Operation Compilation Code
------------------------------------------------------------------------
-- | Compiles an array assignment statement.  The compiled expression
-- and a list of compiled kernels is returned.
compileAssignAExpr
  :: (Scalar a)
  => Id             -- ^  the name of the variable to assign
  -> AExpr a        -- ^  the array expression to compile
  -> Compile (Statement, [Kernel])
compileAssignAExpr outName expr = do
  aliasInfo <- aexprAliases expr
  case expr of
    AVar n           -> compileArrayVar outName n
    ASliceVar si n   -> compileASliceVar outName si n
    AMap f a         -> let rhsLen = alength a
                            rhsFun idx = f (a ! idx)
                            dims = gridDims1D a
                        in compileAMap' outName rhsLen rhsFun dims aliasInfo
    AMap2 f a b      -> let rhsLen = alength a
                            rhsFun idx = f (a ! idx) (b ! idx)
                            dims = gridDims1D a
                        in compileAMap' outName rhsLen rhsFun dims aliasInfo
    AMap3 f a b c    -> let rhsLen = alength a
                            rhsFun idx = f (a ! idx) (b ! idx) (c ! idx)
                            dims = gridDims1D a
                        in compileAMap' outName rhsLen rhsFun dims aliasInfo

compileArrayVar :: Id -> Id -> Compile (Statement, [Kernel])
compileArrayVar o v =
  return (Statement $ funcall "gpu_array_copy" [var v, var o] <> semi, [])

compileASliceInfo
  :: ASliceInfo
  -> Compile ((Doc, Doc, Doc), Statement, [Kernel])
compileASliceInfo (Slice b e s) = runCompileScalar $
  liftM3 (,,) (compileSExpr b) (compileSExpr e) (compileSExpr s)

compileASliceVar
  :: Id
  -> ASliceInfo
  -> Id
  -> Compile (Statement, [Kernel])
compileASliceVar outName sliceInfo varName = do
  ((b', e', s'), hoisted, kernels) <- compileASliceInfo sliceInfo
  let args = [var varName, b', e', s', var outName]
  let copy = Statement $ funcall "array_slice_copy" args <> semi
  return (hoisted `mappend` copy, kernels)

--------------------------------------------------
-- Array map compilation code
--------------------------------------------------
makeArrayMapIndex :: Compile (Id, Statement, SExpr Int)
makeArrayMapIndex = do
  idxName <- uniqueId "i"
  let idxDecl = tInt <+> assign (var idxName) rhs
          where rhs = var "blockIdx.x" <+>
                      char '*' <+>
                      int blockDims <+>
                      char '+' <+>
                      var "threadIdx.x"
  return (idxName, Statement idxDecl, SVar idxName)

compileLoadSharedMemoryArray :: AliasInfo -> Compile Statement
compileLoadSharedMemoryArray aliasInfo = do
  useShmem <- useSharedMemory
  if useShmem
     then do let candidates = Map.keys (arrayCandidates aliasInfo)
             ((), hs) <- runWriterT $ do
                             mapM_ go candidates
                             tell (Statement $ text "__syncthreads();")
             return hs
     else return mempty

  where 
    go :: (Id, SomeScalarWitness) -> CompileKernelScalar ()
    go (n, w) = do
        let nShared = n ++ "_shared"
        let sType = case w of
                      SomeScalarWitness sw -> gpuScalarType sw
        let sDecl = qShared <+> sType <+> var nShared <+>
                    brackets (int blockDims) <> semi
        tell (Statement sDecl)
        let idx = text "blockIdx.x *" <+> int blockDims <+>
                  text "+ threadIdx.x"
        let inBounds = idx <+> char '<' <+> method (var n) "length" []
        let assignment = assign (index (var nShared)
                                       (text "threadIdx.x"))
                                (index (var n) idx)
        tell $ Statement $ ifBlock inBounds assignment


compileLoadSharedMemoryMatrix :: AliasInfo -> Compile Statement
compileLoadSharedMemoryMatrix aliasInfo = do
  useShmem <- useSharedMemory
  if useShmem
     then do let candidates = Map.keys (matrixCandidates aliasInfo)
             ((), hs) <- runWriterT $ do
                             mapM_ go candidates
                             tell (Statement $ text "__syncthreads();")
             return hs
     else return mempty

  where 
    go :: (Id, SomeScalarWitness) -> CompileKernelScalar ()
    go (n, w) = do
        let nShared = n ++ "_shared"
        let sType = case w of
                      SomeScalarWitness sw -> gpuScalarType sw
        let sDecl = qShared <+> sType <+> var nShared <+>
                    brackets (int blockDimsY) <>
                    brackets (int blockDimsX) <> semi
        tell (Statement sDecl)
        let idxX = text "blockIdx.x *" <+> int blockDimsX <+>
                   text "+ threadIdx.x"
        let idxY = text "blockIdx.y *" <+> int blockDimsY <+>
                   text "+ threadIdx.y"
        let inBoundsX = idxX <+> char '<' <+> method (var n) "cols" []
        let inBoundsY = idxY <+> char '<' <+> method (var n) "rows" []
        let inBounds = inBoundsX <+> text "&&" <+> inBoundsY
        let assignment = assign (var nShared <>
                                 brackets (text "threadIdx.y") <>
                                 brackets (text "threadIdx.x"))
                                (funcall n [idxY, idxX])
        tell $ Statement $ ifBlock inBounds assignment

        


makeArrayMapKernel
  :: (Scalar a)
  => Id
  -> SExpr Int
  -> (SExpr Int -> SExpr a)
  -> AliasInfo
  -> Compile (Id, Kernel, [Arg], Arg)
makeArrayMapKernel outName rhsLen rhsFun aliasInfo = do
  shmemLoad <- compileLoadSharedMemoryArray aliasInfo  
  (idxName, idxDecl, idxVar) <- makeArrayMapIndex
  let rhs = rhsFun idxVar
  -- TODO: if it is safe to pull the hoisted portion out of the kernel, do so.
  (assignment, hoisted) <- runWriterT $ do
      rhs'  <- compileKernelSExpr aliasInfo rhs
      test' <- compileKernelSExpr aliasInfo (idxVar <* rhsLen)
      return $ Statement $
               ifBlock test' $
               assign (index (var outName) (var idxName)) rhs'
  let body = mconcat [shmemLoad, idxDecl, hoisted, assignment]
  inputs <- getNonCudaFreeVars [idxName] rhs
  let output = AArg (sexprType rhs) outName
  (kernelName, kernel) <- kernelFunction inputs output body
  return (kernelName, kernel, inputs, output)

makeArrayMapInvocation
  :: SExpr Int
  -> Id
  -> [Arg]
  -> Arg
  -> Compile (Statement, [Kernel])
makeArrayMapInvocation dims kernelName kInputs kOutput = do
  (gdx, gdxh, gdxks) <- runCompileScalar (compileSExpr dims)
  let gd = (gdx, int 1, int 1)
  let bd = (int blockDims, int 1, int 1)
  invocation <- invokeKernel kernelName kInputs kOutput bd gd
  return (gdxh `mappend` invocation, gdxks)

-- | Compiles a use of AMap to invoke a CUDA kernel.  There is no need
-- to worry about handling nested uses of AMap, as such things are
-- impossible to express due to the structure of the ASTs.
compileAMap'
  :: (Scalar a)
  => Id                     -- ^ output name
  -> SExpr Int              -- ^ length expression
  -> (SExpr Int -> SExpr a) -- ^ indexing function
  -> SExpr Int              -- ^ grid dimensions
  -> AliasInfo              -- ^ array overlap information
  -> Compile (Statement, [Kernel])
compileAMap' outName rhsLen rhsFun dims aliasInfo = do
  (kn, k, is, o) <- makeArrayMapKernel outName rhsLen rhsFun aliasInfo
  (invocation, hoistedKernels) <- makeArrayMapInvocation dims kn is o
  return (invocation, hoistedKernels ++ [k])

--------------------------------------------------
-- Matrix expression compilation code
--------------------------------------------------
compileAssignMExpr
  :: (Scalar a)
  => Id             -- ^  the name of the variable to assign
  -> MExpr a        -- ^  the matrix expression to compile
  -> Compile (Statement, [Kernel])
compileAssignMExpr outName expr = do
  aliasInfo <- mexprAliases expr
  case expr of
    MVar n           -> compileMVar outName n
    MSliceVar si n   -> compileMSliceVar outName si n
    MMap f a         -> compileMMap outName aliasInfo f a
    MMap2 f a b      -> compileMMap2 outName aliasInfo f a b

compileMVar :: Id -> Id -> Compile (Statement, [Kernel])
compileMVar o v =
  return (Statement $ funcall "gpu_matrix_copy" [var v, var o] <> semi, [])

compileMSliceVar :: Id -> MSliceInfo -> Id -> Compile (Statement, [Kernel])
compileMSliceVar outName sliceInfo varName =  do
  ((bi, ei, si), iHoisted, iKernels) <- compileASliceInfo (rowSlice sliceInfo)
  ((bj, ej, sj), jHoisted, jKernels) <- compileASliceInfo (colSlice sliceInfo)
  let args = [var varName, bi, ei, si, bj, ej, sj, var outName]
  let copy = Statement $ funcall "matrix_slice_copy" args <> semi
  let hoisted = iHoisted `mappend` jHoisted
  let kernels = iKernels `mappend` jKernels
  return (hoisted `mappend` copy, kernels)

--------------------------------------------------
-- Matrix map compilation code
--------------------------------------------------
makeMatrixMapRowIndex :: Compile (Id, Statement, SExpr Int)
makeMatrixMapRowIndex = do
  idxName <- uniqueId "y"
  let idxDecl = tInt <+> assign (var idxName) rhs
          where rhs = var "blockIdx.y" <+>
                      char '*' <+>
                      int blockDimsY <+>
                      char '+' <+>
                      var "threadIdx.y"
  return (idxName, Statement idxDecl, SVar idxName)

makeMatrixMapColIndex :: Compile (Id, Statement, SExpr Int)
makeMatrixMapColIndex = do
  idxName <- uniqueId "x"
  let idxDecl = tInt <+> assign (var idxName) rhs
          where rhs = var "blockIdx.x" <+>
                      char '*' <+>
                      int blockDimsX <+>
                      char '+' <+>
                      var "threadIdx.x"
  return (idxName, Statement idxDecl, SVar idxName)

makeMatrixMapKernel
  :: (Scalar a)
  => Id
  -> SExpr Int
  -> SExpr Int
  -> (SExpr Int -> SExpr Int -> SExpr a)
  -> AliasInfo
  -> Compile (Id, Kernel, [Arg], Arg)
makeMatrixMapKernel outName rhsRows rhsCols rhsFun aliasInfo = do
  shmemLoad <- compileLoadSharedMemoryMatrix aliasInfo  
  (xName, xDecl, xVar) <- makeMatrixMapColIndex
  (yName, yDecl, yVar) <- makeMatrixMapRowIndex
  let rhs = rhsFun yVar xVar
  (assignment, hoisted) <- runWriterT $ do
      rhs'  <- compileKernelSExpr aliasInfo rhs
      test' <- compileKernelSExpr aliasInfo
                 (yVar <* rhsRows &&* xVar <* rhsCols)
      let lhs = funcall outName [var yName, var xName]
      return $ Statement $ ifBlock test' $ assign lhs rhs'
  let body = mconcat [shmemLoad, xDecl, yDecl, hoisted, assignment]
  inputs <- getNonCudaFreeVars [xName, yName] rhs
  let output = MArg (sexprType rhs) outName
  (kernelName, kernel) <- kernelFunction inputs output body
  return (kernelName, kernel, inputs, output)

makeMatrixMapInvocation
  :: SExpr Int
  -> SExpr Int
  -> Id
  -> [Arg]
  -> Arg
  -> Compile (Statement, [Kernel])
makeMatrixMapInvocation gridRows gridCols kernelName kInputs kOutput = do
  (gdx, gdxh, gdxks) <- runCompileScalar (compileSExpr gridCols)
  (gdy, gdyh, gdyks) <- runCompileScalar (compileSExpr gridRows)
  let gds = (gdx, gdy, int 1)
  let bds = (int blockDimsX, int blockDimsY, int 1)
  invocation <- invokeKernel kernelName kInputs kOutput bds gds
  return (mconcat [gdxh, gdyh, invocation], gdxks ++ gdyks)


compileMMap
  :: (Scalar a, Scalar b)
  => Id                             -- ^  name of output variable
  -> AliasInfo
  -> (SExpr a -> SExpr b)           -- ^  mapping function
  -> MExpr a                        -- ^  matrix expression to map over
  -> Compile (Statement, [Kernel])
compileMMap outName aliasInfo fun mexpr = do
  let rhsRows = mrows mexpr
      rhsCols = mcols mexpr
      rhsFun i j = fun (mexpr ! (i, j))
      gridCols = (mcols mexpr + (blockDimsX - 1)) `idiv` blockDimsX
      gridRows = (mrows mexpr + (blockDimsY - 1)) `idiv` blockDimsY
  (kn, k, is, o) <- makeMatrixMapKernel outName rhsRows rhsCols rhsFun aliasInfo
  (invocation, hoistedKernels) <- makeMatrixMapInvocation gridRows gridCols kn is o
  return (invocation, hoistedKernels ++ [k])

compileMMap2
  :: (Scalar a, Scalar b, Scalar c)
  => Id                               -- ^  name of output variable
  -> AliasInfo
  -> (SExpr a -> SExpr b -> SExpr c)  -- ^  mapping function
  -> MExpr a                          -- ^  first matrix expression to map over
  -> MExpr b                          -- ^  second matrix expression to map over
  -> Compile (Statement, [Kernel])
compileMMap2 outName aliasInfo fun mexpr1 mexpr2 = do
  let rhsRows = mrows mexpr1
      rhsCols = mcols mexpr1
      rhsFun i j = fun (mexpr1 ! (i, j)) (mexpr2 ! (i, j))
      gridCols = (mcols mexpr1 + (blockDimsX - 1)) `idiv` blockDimsX
      gridRows = (mrows mexpr1 + (blockDimsY - 1)) `idiv` blockDimsY
  (kn, k, is, o) <- makeMatrixMapKernel outName rhsRows rhsCols rhsFun aliasInfo
  (invocation, hoistedKernels) <- makeMatrixMapInvocation gridRows gridCols kn is o
  return (invocation, hoistedKernels ++ [k])

--------------------------------------------------
-- Common CUDA kernel compilation code
--------------------------------------------------
dim3 :: Id -> (Doc, Doc, Doc) -> Doc
dim3 n (x, y, z) = var n <+> parens (commaSep [x, y, z])

blockDims :: (Num a) => a
blockDims = 256

blockDimsX :: (Num a) => a
blockDimsX = 16

blockDimsY :: (Num a) => a
blockDimsY = 16

gridDims1D :: (Scalar a) => AExpr a -> SExpr Int
gridDims1D aexpr = (alength aexpr + (blockDims - 1)) `idiv` blockDims

-- | CUDA's special variables.
cudaVars :: [Id]
cudaVars = roots ++ specials
  where
    specials = [ v ++ "." ++ m | v <- roots, m <- members ]
    roots = [ "blockIdx"
            , "gridIdx"
            , "threadIdx"
            , "blockDim"
            , "gridDim"
            ]
    members = ["x", "y", "z"]

getNonCudaFreeVars :: (Scalar a) => [Id] -> SExpr a -> Compile [Arg]
getNonCudaFreeVars bound expr = liftM (filter notSpecial) vars
  where notSpecial a = argName a `notElem` cudaVars
        vars = getFreeVars bound expr

-- | Generate code for a kernel invocation.
invokeKernel
  :: Id                 -- ^  kernel name
  -> [Arg]              -- ^  arguments
  -> Arg                -- ^  output
  -> (Doc, Doc, Doc)    -- ^  block dimensions
  -> (Doc, Doc, Doc)    -- ^  grid dimensions
  -> Compile Statement
invokeKernel kernelName args output bDims gDims = do
  let bracks d = text "<<<" <> d <> text ">>>"
  bDimsName <- uniqueId "blockDims"
  gDimsName <- uniqueId "gridDims"
  let binds = qConst <+> text "dim3" <+> dim3 bDimsName bDims <> semi $+$
              qConst <+> text "dim3" <+> dim3 gDimsName gDims <> semi
  let kparams = bracks (commaSep [var gDimsName, var bDimsName])
  let params = kernelActualParamList args output
  let invocation = text kernelName <> kparams <> params <> semi
  return (Statement binds `mappend` Statement invocation)

kernelActualParamList :: [Arg] -> Arg -> Doc
kernelActualParamList is o = parens (commaSep paramList)
  where
    paramList = map actualParam is ++ [actualOutputParam o]

    -- | Convert an Arg into a list of values to supply to a kernel.
    actualParam :: Arg -> Doc
    actualParam arg =
      case arg of
        SArg _sw n -> var n
        AArg sw n  -> kernelArrayType sw <+> parens (var n)
        MArg sw n  -> kernelMatrixType sw <+> parens (var n)

    actualOutputParam :: Arg -> Doc
    actualOutputParam arg =
      case arg of
        SArg {} -> error "kernelActualOutputParam: scalar output parameter?"
        AArg sw n -> kernelOutputArrayType sw <+> parens (var n)
        MArg sw n -> kernelOutputMatrixType sw <+> parens (var n)

-- | Create a kernel function.
kernelFunction
  :: [Arg]      -- ^  input arguments
  -> Arg        -- ^  output argument
  -> Statement  -- ^  body
  -> Compile (Id, Kernel)
kernelFunction inputs output body = do
  kernelName <- uniqueId "kernel"
  let kernel = qStatic <+> qGlobal <+> tVoid $+$
               var kernelName <+> kernelArgList inputs output $+$
               block (unStatement body)
  return (kernelName, Kernel kernel)
