{-# LANGUAGE GADTs #-}
module Language.Barracuda.Compiler.FreeVariables
  ( getFreeVars
  ) where

import Language.Barracuda.AST
import Language.Barracuda.Compiler.Infrastructure
import Language.Barracuda.Function

import Control.Monad (liftM, liftM2)
import Data.List (nub)

-- TODO: `getFreeVars' should be smarter, figuring out which array
-- lengths are needed, so that the length of every array expression
-- will not need to be passed as an argument.
-- | Find all the free
-- variables within a scalar expression given a list of binding names.
getFreeVars
  :: (Scalar a)
  => [Id]           -- ^  list of bound variables
  -> SExpr a        -- ^  expression to inspect
  -> Compile [Arg]
getFreeVars boundNames = liftM nub . sexprFreeVars
  where
    sexprFreeVars :: SExpr a -> Compile [Arg]
    sexprFreeVars expr =
      case expr of
        SLet b f -> do b' <- getFreeVars boundNames b
                       bName <- uniqueId "let"
                       f' <- getFreeVars (bName : boundNames) (f $ SVar bName)
                       return (b' ++ f')
        SVar n -> return $ if n `elem` boundNames
                           then []
                           else [SArg (sexprType expr) n]
        SConst _c -> return []
        Prim1 _op e -> sexprFreeVars e
        Prim2 _op e1 e2 -> liftM2 (++) (sexprFreeVars e1) (sexprFreeVars e2)
        SCond p t f -> do p' <- sexprFreeVars p
                          t' <- sexprFreeVars t
                          f' <- sexprFreeVars f
                          return (p' ++ t' ++ f')
        AReduce f i t -> do accName <- uniqueId "acc"
                            valName <- uniqueId "val"
                            f' <- getFreeVars (accName : valName : boundNames)
                                              (f (SVar accName) (SVar valName))
                            i' <- sexprFreeVars i
                            t' <- aexprFreeVars t
                            return (f' ++ i' ++ t')
        AIdx n i -> do let n' = if n `elem` boundNames
                                then []
                                else [AArg (sexprType expr) n]
                       i' <- sexprFreeVars i
                       return (n' ++ i')
        ALen t -> aexprFreeVars t
        MReduce f i m -> do accName <- uniqueId "acc"
                            valName <- uniqueId "val"
                            f' <- getFreeVars (accName : valName : boundNames)
                                              (f (SVar accName) (SVar valName))
                            i' <- sexprFreeVars i
                            m' <- mexprFreeVars m
                            return (f' ++ i' ++ m')
        MIdx n i j -> do let n' = if n `elem` boundNames
                                  then []
                                  else [MArg (sexprType expr) n]
                         i' <- sexprFreeVars i
                         j' <- sexprFreeVars j
                         return (n' ++ i' ++ j')
        MRows m -> mexprFreeVars m
        MCols m -> mexprFreeVars m

    aexprFreeVars :: AExpr a -> Compile [Arg]
    aexprFreeVars aexpr =
      case aexpr of
        AVar n -> if n `elem` boundNames
                  then return []
                  else return [AArg (aexprElemType aexpr) n]
        ASliceVar (Slice b e s) v -> do
                   b' <- sexprFreeVars b
                   e' <- sexprFreeVars e
                   s' <- sexprFreeVars s
                   let v' = if v `elem` boundNames
                            then []
                            else [AArg (aexprElemType aexpr) v]
                   return (b' ++ e' ++ s' ++ v')
        AMap f a -> do
          valName <- uniqueId "val"
          f' <- getFreeVars (valName : boundNames) (f (SVar valName))
          a' <- aexprFreeVars a
          return (f' ++ a')
        AMap2 f a b -> do
          val1Name <- uniqueId "val"
          val2Name <- uniqueId "val"
          f' <- getFreeVars (val1Name : val2Name: boundNames)
                            (f (SVar val1Name) (SVar val2Name))
          a' <- aexprFreeVars a
          b' <- aexprFreeVars b
          return (f' ++ a' ++ b')
        AMap3 f a b c -> do
          val1Name <- uniqueId "val"
          val2Name <- uniqueId "val"
          val3Name <- uniqueId "val"
          f' <- getFreeVars (val1Name : val2Name: boundNames)
                            (f (SVar val1Name) (SVar val2Name) (SVar val3Name))
          a' <- aexprFreeVars a
          b' <- aexprFreeVars b
          c' <- aexprFreeVars c
          return (f' ++ a' ++ b' ++ c')

    mexprFreeVars :: MExpr a -> Compile [Arg]
    mexprFreeVars mexpr =
      case mexpr of
        MVar n -> if n `elem` boundNames
                  then return []
                  else return [MArg (mexprElemType mexpr) n]
        MSliceVar sliceInfo v -> do
                   let Slice bi ei si = rowSlice sliceInfo
                   let Slice bj ej sj = colSlice sliceInfo
                   bi' <- sexprFreeVars bi
                   ei' <- sexprFreeVars ei
                   si' <- sexprFreeVars si
                   bj' <- sexprFreeVars bj
                   ej' <- sexprFreeVars ej
                   sj' <- sexprFreeVars sj
                   let v' = if v `elem` boundNames
                            then []
                            else [MArg (mexprElemType mexpr) v]
                   return $ concat [bi', ei', si', bj', ej', sj', v']
        MMap f a -> do
          valName <- uniqueId "val"
          f' <- getFreeVars (valName : boundNames) (f (SVar valName))
          a' <- mexprFreeVars a
          return (f' ++ a')
        MMap2 f a b -> do
          val1Name <- uniqueId "val"
          val2Name <- uniqueId "val"
          f' <- getFreeVars (val1Name : val2Name: boundNames)
                            (f (SVar val1Name) (SVar val2Name))
          a' <- mexprFreeVars a
          b' <- mexprFreeVars b
          return (f' ++ a' ++ b')
