{-# LANGUAGE GADTs, KindSignatures, FlexibleContexts, FlexibleInstances, StandaloneDeriving #-}
{-# OPTIONS -Wall -fno-warn-missing-signatures #-}
module Language.Barracuda.Function
  (
    Arg(..)
  , argName
  , SomeExpr(..)
  , Instantiable(..)
  , Function(..)
  , functionName
  ) where

import Language.Barracuda.AST
import Data.Typeable (cast)

-- | Reification of arguments to a Barracuda function.  A witness of the
-- type, along with the variable name, can be obtained from an Arg.
data Arg :: * where
  SArg   :: (Scalar a) => ScalarWitness a -> Id -> Arg
  AArg   :: (Scalar a) => ScalarWitness a -> Id -> Arg
  MArg   :: (Scalar a) => ScalarWitness a -> Id -> Arg

deriving instance Show Arg

-- | Get the name of the argument.
argName :: Arg -> Id
argName (SArg _sw n) = n
argName (AArg _sw n) = n
argName (MArg _sw n) = n

eqTest w1 w2 n1 n2 = maybe False (== w2) (cast w1) && n1 == n2

instance Eq Arg where
  SArg w1 n1 == SArg w2 n2 = eqTest w1 w2 n1 n2
  AArg w1 n1 == AArg w2 n2 = eqTest w1 w2 n1 n2
  MArg w1 n1 == MArg w2 n2 = eqTest w1 w2 n1 n2
  _ == _ = False

-- | An existential wrapper around expressions.
data SomeExpr :: * where
  SExpr :: (Scalar a) => SExpr a -> SomeExpr
  AExpr :: (Scalar a) => AExpr a -> SomeExpr
  MExpr :: (Scalar a) => MExpr a -> SomeExpr

-- | A class of things that can be instantiated with variables.  This
-- is used to fill functions on expressions with variables, and
-- compile them.
--
-- This class is conceptually closed---all the instances for it are
-- defined in this file.  This type class defines a type-indexed
-- recursive function on scalar & array expressions, and functions on
-- scalar & array expressions.  It allows things typed like
--
--     SExpr Float
--     SExpr Float -> SExpr Float
--     SExpr Float -> AExpr Float -> SExpr Float
--
-- to be compiled.
class Instantiable a where
  -- | Takes a list of names to use for input/output variables, the
  -- thing to instantiate, and returns a list of input arguments, the
  -- output argument, and the instantiated result.
  --
  -- The list of names must not be empty.  All the instances of
  -- Instantiable are undefined if the wrong number of names is given.
  instantiate :: [Id] -> a -> ([Arg], Arg, SomeExpr)


-- | Common code used in the `base cases' for the Instantiable
-- instances.
instantiateBasis
  :: (t -> Id -> Arg)            -- ^  a function to create an Arg
  -> (t -> SomeExpr)             -- ^  a function to create a SomeExpr
  -> [Id]                        -- ^  the singleton list of the output name
  -> t                           -- ^  the expression type
  -> ([Arg], Arg, SomeExpr)
instantiateBasis _ _ [] _ = error "instantiateBasis: empty name list"
instantiateBasis argf exprf [out] expr = ([], argf expr out, exprf expr)
instantiateBasis _ _ _ _  = error "instantiateBasis: too many names"

-- The base cases for Instantiable, namely, for scalar and array
-- expressions.
instance (Scalar a) => Instantiable (SExpr a) where
  instantiate = instantiateBasis (SArg . sexprType) SExpr

instance (Scalar a) => Instantiable (AExpr a) where
  instantiate = instantiateBasis (AArg . aexprElemType) AExpr

instance (Scalar a) => Instantiable (MExpr a) where
  instantiate = instantiateBasis (MArg . mexprElemType) MExpr

-- | Common code used in the `inductive cases' for the Instantiable
-- instances.
instantiateInduction
  :: (Instantiable a)
  => (t -> Id -> Arg)             -- ^  a function to create an Arg
  -> (Id -> t)                    -- ^  a function to create a variable
  -> [Id]                         -- ^  the list of input/output names
  -> (t -> a)                     -- ^  the function to instantiate
  -> ([Arg], Arg, SomeExpr)
instantiateInduction _ _ [] _ = error "instantiateInduction: empty name list"
instantiateInduction argf varf (n : ns) f =
  let v = varf n
      f' = f v
      (is, o, e) = instantiate ns f'
      res = (argf v n : is, o, e)
  in res

-- The inductive cases for Instantiable, namely, instances for
-- functions whose arguments are either scalar or array expressions.
instance (Scalar a, Instantiable r) => Instantiable (SExpr a -> r) where
  instantiate = instantiateInduction (SArg . sexprType) SVar

instance (Scalar a, Instantiable r) => Instantiable (AExpr a -> r) where
  instantiate = instantiateInduction (AArg . aexprElemType) AVar

instance (Scalar a, Instantiable r) => Instantiable (MExpr a -> r) where
  instantiate = instantiateInduction (MArg . mexprElemType) MVar

-- | Annotated Barracuda functions.
data Function :: * where
  Function
    :: (Instantiable a)
    => a                -- an Instantiable
    -> Id               -- the name to give the function
    -> [Id]             -- the names to give the input arguments
    -> Id               -- the name to give the output argument
    -> Function

-- | Get the name of a function.
functionName :: Function -> Id
functionName (Function _fun name _inputs _output) = name
