#include "test_array.hpp"
#include "random.hpp"
#include "timer.hpp"
#include "cuda_timer.hpp"
#include "timeout.hpp"

#include <cstdio>
#include <cstdlib>
#include <cassert>
#include <cstring>
#include <cerrno>

#include "gpu_shmem_benchmarks.hpp"
#include "gpu_noshmem_benchmarks.hpp"
#include "gpu_benchmarks.hpp"
#include "cpu_benchmarks.hpp"
#include "rmse_benchmarks.hpp"
#include "saxpy_benchmarks.hpp"
#include "sdot_benchmarks.hpp"
#include "forward_diff_benchmarks.hpp"
#include "black_scholes_benchmarks.hpp"
#include "add_one_2d_benchmarks.hpp"
#include "jacobi_stencil_benchmarks.hpp"
#include "weighted_moving_average_benchmarks.hpp"


namespace
{
    const unsigned MAX_ARRAY_LENGTH = 16777216 / 2;
    const unsigned MAX_MATRIX_DIM = 4096;
    const unsigned TARGET_ELAPSED_SECONDS = 20;
} // end namespace


namespace
{
    void run_array_benchmark (array_benchmark &benchmark)
    {
        FILE *outfile = fopen(benchmark.get_outfile_name(), "w");
        if (outfile == NULL) {
            fprintf (
                stderr,
                "error: could not open file `%s' for writing: %s\n",
                benchmark.get_outfile_name(),
                strerror(errno)
                );
            abort();
        }
        
        printf (
            "%-24s%16s%16s%16s\n",
            "# procedure",
            "total runtime",
            "num. of runs",
            "length"
            );
        fprintf (
            outfile,
            "%-20s%-20s%-20s\n",
            "total_runtime",
            "num_runs",
            "length"
            );

        for (unsigned length = 1; length <= MAX_ARRAY_LENGTH; length *= 2) 
        {
            cuda_timer timer;
            unsigned num_runs = 0;

            benchmark.setup(length);
            cudaThreadSynchronize();

            // warmup
            benchmark.run();
            benchmark.run();
            benchmark.run();
            cudaThreadSynchronize();

            timeout (TARGET_ELAPSED_SECONDS);
            while (keep_running())
            {
                timer.start();
                benchmark.run();
                timer.stop();
                num_runs = num_runs + 1;
            }
            timeout_stop ();

            cudaThreadSynchronize();
            benchmark.teardown();

            printf (
                "%-24s%16.10lf%16u%16u\n",
                benchmark.get_name(),
                timer.get_elapsed_seconds(),
                num_runs,
                length
                );
            fprintf (
                outfile,
                "%-20.10lf%-20u%-20u\n",
                timer.get_elapsed_seconds(),
                num_runs,
                length
                );
        }

        if (fclose(outfile))
        {
            fprintf (
                stderr,
                "error: could not close file `%s': %s\n",
                benchmark.get_outfile_name(),
                strerror(errno)
                );
            abort();
        }
    }

    void run_matrix_benchmark (matrix_benchmark &benchmark)
    {
        FILE *outfile = fopen(benchmark.get_outfile_name(), "w");
        if (outfile == NULL) {
            fprintf (
                stderr,
                "error: could not open file `%s' for writing: %s\n",
                benchmark.get_outfile_name(),
                strerror(errno)
                );
            abort();
        }
        
        printf (
            "%-24s%16s%16s%16s\n",
            "# procedure",
            "total runtime",
            "num. of runs",
            "length"
            );
        fprintf (
            outfile,
            "%-20s%-20s%-20s\n",
            "total_runtime",
            "num_runs",
            "length"
            );

        for (unsigned length = 2; length <= MAX_MATRIX_DIM; length *= 2) 
        {
            cuda_timer timer;
            unsigned num_runs = 0;

            benchmark.setup(length, length);
            cudaThreadSynchronize();

            // warmup
            benchmark.run();
            benchmark.run();
            benchmark.run();
            cudaThreadSynchronize();

            timeout (TARGET_ELAPSED_SECONDS);
            while (keep_running())
            {
                timer.start();
                benchmark.run();
                timer.stop();
                num_runs = num_runs + 1;
            }
            timeout_stop ();

            cudaThreadSynchronize();
            benchmark.teardown();

            printf (
                "%-24s%16.10lf%16u%16u\n",
                benchmark.get_name(),
                timer.get_elapsed_seconds(),
                num_runs,
                length
                );
            fprintf (
                outfile,
                "%-20.10lf%-20u%-20u\n",
                timer.get_elapsed_seconds(),
                num_runs,
                length
                );
        }

        if (fclose(outfile))
        {
            fprintf (
                stderr,
                "error: could not close file `%s': %s\n",
                benchmark.get_outfile_name(),
                strerror(errno)
                );
            abort();
        }
    }
} // end namespace




int main (int argc, char *argv[])
{
    rand_src gen;

    array_benchmark * black_scholes_benchmarks[] = {
        new gpu_black_scholes_benchmark("results/gpu_black_scholes.dat", gen),
        new cuda_sdk_gpu_black_scholes_benchmark("results/cuda_sdk_gpu_black_scholes.dat", gen),
        NULL
    };

    array_benchmark * saxpy_benchmarks[] = {
        new gpu_saxpy_benchmark("results/gpu_saxpy.dat", gen),
        new cublas_saxpy_benchmark("results/cublas_saxpy.dat", gen),
        NULL
    };

    array_benchmark * sdot_benchmarks[] = {
        new gpu_sdot_benchmark("results/gpu_sdot.dat", gen),
        new cublas_sdot_benchmark("results/cublas_sdot.dat", gen),
        NULL
    };

    array_benchmark * forward_diff_benchmarks[] = {
        new gpu_shmem_forward_diff_benchmark("results/gpu_forward_diff_shmem.dat", gen),
        new gpu_noshmem_forward_diff_benchmark("results/gpu_forward_diff_noshmem.dat", gen),
        NULL
    };

    array_benchmark * rmse_benchmarks[] = {
        new gpu_rmse_fused_benchmark("results/gpu_rmse_fused.dat", gen),
        new gpu_rmse_unfused_benchmark("results/gpu_rmse_unfused.dat", gen),
        NULL
    };

    array_benchmark * weighted_moving_average_benchmarks[] = {
        new gpu_shmem_weighted_moving_average_benchmark("results/gpu_weighted_moving_average_shmem.dat", gen),
        new gpu_noshmem_weighted_moving_average_benchmark("results/gpu_weighted_moving_average_noshmem.dat", gen),
        NULL
    };

    array_benchmark ** array_suites[] = {
        saxpy_benchmarks,
        sdot_benchmarks,
        forward_diff_benchmarks,
        rmse_benchmarks,
        weighted_moving_average_benchmarks,
        black_scholes_benchmarks,
        NULL
    };

    matrix_benchmark * jacobi_stencil_benchmarks[] = {
        new gpu_noshmem_jacobi_stencil_benchmark ("results/gpu_noshmem_jacobi_stencil.dat", gen),
        new gpu_shmem_jacobi_stencil_benchmark ("results/gpu_shmem_jacobi_stencil.dat", gen),
        new handwritten_gpu_noshmem_jacobi_stencil_benchmark ("results/handwritten_gpu_noshmem_jacobi_stencil.dat", gen),
        new handwritten_gpu_shmem_jacobi_stencil_benchmark ("results/handwritten_gpu_shmem_jacobi_stencil.dat", gen),
        NULL
    };

    matrix_benchmark * add_one_2d_benchmarks[] = {
        new gpu_add_one_2d_benchmark("results/gpu_add_one_2d.dat", gen),
        NULL
    };

    matrix_benchmark ** matrix_suites[] = {
        jacobi_stencil_benchmarks,
        add_one_2d_benchmarks,
        NULL
    };

    for (unsigned s = 0; array_suites[s] != NULL; s += 1)
        for (unsigned i = 0; array_suites[s][i] != NULL; i += 1)
            run_array_benchmark (*array_suites[s][i]);

    for (unsigned s = 0; matrix_suites[s] != NULL; s += 1)
        for (unsigned i = 0; matrix_suites[s][i] != NULL; i += 1)
            run_matrix_benchmark (*matrix_suites[s][i]);

    for (unsigned s = 0; array_suites[s] != NULL; s += 1)
        for (unsigned i = 0; array_suites[s][i] != NULL; i += 1)
            delete array_suites[s][i];

    for (unsigned s = 0; matrix_suites[s] != NULL; s += 1)
        for (unsigned i = 0; matrix_suites[s][i] != NULL; i += 1)
            delete matrix_suites[s][i];

    return 0;
}
