module Main (main) where

import Language.Barracuda


main :: IO ()
main = do
  compileFunctions [ saxpyFun
                   , sdotFun
                   , blackScholesFun
                   , addOne2DFun
                   , jacobiStencilFun
                   , wmAvgFun
                   ]

  compile defaultCudaConfig
          ("gpu_rmse_benchmarks.hpp", "gpu_rmse_benchmarks.cu")
          rmseFuns

  let gpuFuns = [forwardDiffFun, jacobiStencilFun, wmAvgFun]
  let shmemConf = CudaTarget CudaConfig { sharedMemoryOptimization = True }
  let noshmemConf = defaultCudaConfig
  compile shmemConf ("gpu_shmem_benchmarks.hpp", "gpu_shmem_benchmarks.cu")
              (map ($ "gpu_shmem") gpuFuns)
  compile noshmemConf ("gpu_noshmem_benchmarks.hpp", "gpu_noshmem_benchmarks.cu")
              (map ($ "gpu_noshmem") gpuFuns)

-- | Compile a function for both CPU and GPU, with shared memory
-- optimization disabled.
compileFunctions
  :: [String -> Function]  -- ^ Functions that give a Function, given a prefix
  -> IO ()
compileFunctions fs = do
  let cpuFuns  = map ($ "cpu") fs
  let gpuFuns  = map ($ "gpu") fs
  let cpuConf = defaultCppConfig
  let gpuConf = defaultCudaConfig
  compile cpuConf ("cpu_benchmarks.hpp", "cpu_benchmarks.cpp") cpuFuns
  compile gpuConf ("gpu_benchmarks.hpp", "gpu_benchmarks.cu") gpuFuns


-- | SAXPY from BLAS Level 1.
saxpy :: SExpr Float -> AExpr Float -> AExpr Float -> AExpr Float
saxpy a x y = azipWith (+) (amap (* a) x) y

saxpyFun :: String -> Function
saxpyFun p = Function saxpy (p ++ "_saxpy") ["alpha", "x", "y"] "result"

-- | SDOT from BLAS Level 1
sdot :: AExpr Float -> AExpr Float -> SExpr Float
sdot x y = asum $ azipWith (*) x y

sdotFun :: String -> Function
sdotFun p = Function sdot (p ++ "_sdot") ["x", "y"] "result"

-- | Forward difference of an array.
forwardDiff :: AExpr Float -> AExpr Float
forwardDiff x = azipWith (-) x'' x'
  where x'  = aslice x (0, alength x - 1, 1)
        x'' = aslice x (1, alength x, 1)

forwardDiffFun :: String -> Function
forwardDiffFun p = Function forwardDiff (p ++ "_forward_diff") ["x"] "result"

-- | Black-Scholes call option pricing.  Riffed from the Nikola source.
blackScholes :: AExpr Float
             -> AExpr Float
             -> AExpr Float
             -> SExpr Float
             -> SExpr Float
             -> AExpr Float
blackScholes ss xs ts r v =
    azipWith3 (\s x t -> blackScholes' True s x t r v) ss xs ts
  where
    blackScholes' :: Bool        -- ^ @True@ for call, @False@ for put
                  -> SExpr Float -- ^ Stock price
                  -> SExpr Float -- ^ Option strike
                  -> SExpr Float -- ^ Option years
                  -> SExpr Float -- ^ Riskless rate
                  -> SExpr Float -- ^ Volatility rate
                  -> SExpr Float
    blackScholes' isCall s x t r v | isCall    = call
                                   | otherwise = put
      where
        call = slet (v * sqrt t) $ \vsqrtt ->
               slet ((log(s/x) + (r+0.5*v*v)*t)/vsqrtt) $ \d1 ->
               slet (d1 - vsqrtt) $ \d2 ->
               s * normcdf d1 - x*exp (-r*t) * normcdf d2
        put  = slet (v * sqrt t) $ \vsqrtt ->
               slet ((log(s/x) + (r+0.5*v*v)*t)/vsqrtt) $ \d1 ->
               slet (d1 - vsqrtt) $ \d2 ->
               x * exp (-r*t) * normcdf (-d2) - s * normcdf (-d1)
    
    normcdf :: SExpr Float -> SExpr Float
    normcdf d = normcdf'
      where
        normcdf' = 
            slet (abs d) $ \l ->
            slet (1.0 / (1.0 + 0.2316419 * l)) $ \k ->
            slet (1.0 / sqrt (2.0 * pi) * exp(- 0.5 * d * d) * poly k) $ \cnd ->
            scond (d >* 0) (1 - cnd) cnd
        poly v = slet v $ \v' -> horner coeff v'
        coeff = [0.0,0.31938153,-0.356563782,1.781477937,-1.821255978,1.330274429]
    
    horner :: Num a => [a] -> a -> a
    horner coeff x = foldr1 madd coeff
      where
        madd a b = b*x + a
    
blackScholesFun :: String -> Function
blackScholesFun p = Function blackScholes funName funParams "result"
  where funName = p ++ "_black_scholes"
        funParams = ["s", "x", "t", "r", "v"]

-- | One step of Jacobi iteration
jacobiStep :: MExpr Float -> MExpr Float
jacobiStep m = mmap (/ 4) (n .+. s .+. e .+. w)
  where
    (.+.) = mzipWith (+)
    n = mslice m (0, mrows m - 2, 1) (1, mcols m - 1, 1)
    s = mslice m (2, mrows m, 1)     (1, mcols m - 1, 1)
    e = mslice m (1, mrows m - 1, 1) (2, mcols m, 1)
    w = mslice m (1, mrows m - 1, 1) (0, mcols m - 2, 1)

jacobiStencilFun :: String -> Function
jacobiStencilFun p = Function jacobiStep funName ["input"] "result"
  where funName = p ++ "_jacobi_stencil"

addOne2D :: MExpr Float -> MExpr Float
addOne2D = mmap (+ 1)

addOne2DFun :: String -> Function
addOne2DFun p = Function addOne2D (p ++ "_add_one_2d") ["input"] "result"

wmAvg :: AExpr Float -> AExpr Float
wmAvg xs = foldr1 (azipWith (+)) slices
  where
    slices :: [AExpr Float]
    slices = zipWith weight (map int [0..]) spencerWeights

    weight :: SExpr Int -> SExpr Float -> AExpr Float
    weight i w = amap (* w) (aslice xs (i, sliceLen + i, 1))

    sliceLen :: SExpr Int
    sliceLen = alength xs - int (length spencerWeights)

    spencerWeights :: [SExpr Float]
    spencerWeights = [-3, -6, -5, 3, 21, 46, 67, 74, 67, 46, 21, 3, -5, -6, -3]

wmAvgFun :: String -> Function
wmAvgFun p = Function wmAvg funName ["input"] "output"
  where funName = p ++ "_weighted_moving_average"

rmse :: AExpr Float -> AExpr Float -> SExpr Float
rmse xs ys = sqrt (sumDiff / len)
  where
    len = intToFloat (alength xs)
    sumDiff = asum (amap (^2) (azipWith (-) xs ys))

rmseFuns :: [Function]
rmseFuns = gpu_rmse : rmseSlow
  where gpu_rmse = Function rmse "gpu_rmse" ["xs", "ys"] "output"

rmseSlow :: [Function]
rmseSlow =
    [ Function rmseSub "rmseSub" ["xs", "ys"] "output"
    , Function rmsePow "rmsePow"  ["xs"] "output"
    , Function rmseSum "rmseSum" ["xs"] "output"
    ]
  where
    rmseSub :: AExpr Float -> AExpr Float -> AExpr Float
    rmseSub xs ys = azipWith (-) xs ys

    rmsePow :: AExpr Float -> AExpr Float
    rmsePow xs = amap (^2) xs

    rmseSum :: AExpr Float -> SExpr Float
    rmseSum xs = asum xs
