#ifndef BARRACUDA_MATRIX_HPP
#define BARRACUDA_MATRIX_HPP

#include <cassert>
#include <algorithm>

#ifndef NO_CUDA
template <class T>
class gpu_matrix;
#endif

#include "host_and_device.hpp"
#include "utility_functions.hpp"

template <class T>
class matrix
{
public:
  // -----------------------------------------------------------
  // Constructors / Destructor
  // -----------------------------------------------------------
  matrix(const int rows, const int cols)
    : _rows(rows)
    , _cols(cols)
    , _store(new T[_rows * _cols])
  {
  }

  matrix(const int rows, const int cols, const T *values)
    : _rows(rows)
    , _cols(cols)
    , _store(new T[_rows * _cols])
  {
    std::copy(values, values + _rows * _cols, _store);
  }

#ifndef NO_CUDA
  matrix(const gpu_matrix<T> &other)
    : _rows(other.rows())
    , _cols(other.cols())
    , _store(new T[_rows * _cols])
  {
    const size_t width = _cols * sizeof(T);
    const size_t height = _rows;
    const size_t pitch = width;
    cudaMemcpy2D(_store,
                 pitch,
                 other.raw_matrix(),
                 other.pitch(),
                 width,
                 height,
                 cudaMemcpyDeviceToHost);
  }
#endif

  ~matrix()
  {
    delete [] _store;
  }

  // -----------------------------------------------------------
  // Indexing
  // -----------------------------------------------------------
  const T & operator ()(int i, int j) const
  {
    assert(i < _rows);
    assert(j < _cols);
    return _store[i * _cols + j];
  }

  T & operator ()(int i, int j)
  {
    assert(i < _rows);
    assert(j < _cols);
    return _store[i * _cols + j];
  }

  // -----------------------------------------------------------
  // Accessors
  // -----------------------------------------------------------
  int rows() const
  {
    return _rows;
  }

  int cols() const
  {
    return _cols;
  }

private:
  const int _rows;
  const int _cols;
  T *_store;
};

#endif
