module Main ( main ) where

import Language.Barracuda

blackScholes :: AExpr Float
             -> AExpr Float
             -> AExpr Float
             -> SExpr Float
             -> SExpr Float
             -> AExpr Float
blackScholes ss xs ts r v =
    azipWith3 (\s x t -> blackScholes' True s x t r v) ss xs ts
  where
    blackScholes' :: Bool        -- ^ @True@ for call, @False@ for put
                  -> SExpr Float -- ^ Stock price
                  -> SExpr Float -- ^ Option strike
                  -> SExpr Float -- ^ Option years
                  -> SExpr Float -- ^ Riskless rate
                  -> SExpr Float -- ^ Volatility rate
                  -> SExpr Float
    blackScholes' isCall s x t r v | isCall    = call
                                   | otherwise = put
      where
        call = slet (v * sqrt t) $ \vsqrtt ->
               slet ((log(s/x) + (r+0.5*v*v)*t)/vsqrtt) $ \d1 ->
               slet (d1 - vsqrtt) $ \d2 ->
               s * normcdf d1 - x*exp (-r*t) * normcdf d2
        put  = slet (v * sqrt t) $ \vsqrtt ->
               slet ((log(s/x) + (r+0.5*v*v)*t)/vsqrtt) $ \d1 ->
               slet (d1 - vsqrtt) $ \d2 ->
               x * exp (-r*t) * normcdf (-d2) - s * normcdf (-d1)
    
    normcdf :: SExpr Float -> SExpr Float
    normcdf d = normcdf'
      where
        normcdf' = 
            slet (abs d) $ \l ->
            slet (1.0 / (1.0 + 0.2316419 * l)) $ \k ->
            slet (1.0 / sqrt (2.0 * pi) * exp(- 0.5 * d * d) * poly k) $ \cnd ->
            scond (d >* 0) (1 - cnd) cnd
        poly v = slet v $ \v' -> horner coeff v'
        coeff = [0.0,0.31938153,-0.356563782,1.781477937,-1.821255978,1.330274429]
    
    horner :: Num a => [a] -> a -> a
    horner coeff x = foldr1 madd coeff
      where
        madd a b = b*x + a

bsTest :: String -> [Function]
bsTest p = [Function blackScholes funName funParams "result"]
  where funName = p ++ "_black_scholes"
        funParams = ["s", "x", "t", "r", "v"]

main :: IO ()
main = do
  compile defaultCppConfig("cpu_black_scholes.hpp", "cpu_black_scholes.cpp") (bsTest "cpu")
  compile defaultCudaConfig ("gpu_black_scholes.hpp", "gpu_black_scholes.cu") (bsTest "gpu")
