#include "barracuda_test.hpp"
#include "test_cases.hpp"

#include <cstdio>
#include <cassert>


namespace {
  // const int MAX_ARRAY_LEN = 16777216;  // 2^24
  const int MAX_ARRAY_LEN = 8388608;   // 2^23
  // const int MAX_ARRAY_LEN = 4194304;  // 2^22

  const int MAX_MATRIX_ROWS = 32768;
  const int MAX_MATRIX_COLS = 32768;
  const int MAX_NUM_ELEMS = MAX_ARRAY_LEN * 2;

  const int FUZZ_START = -2;
  const int FUZZ_STOP = 2;


  void run_array_test(float_gen &generator,
                      const char *name,
                      bool test(float_gen &, const unsigned))
  {
    for (int base_len = 1; base_len <= MAX_ARRAY_LEN; base_len *= 2)
      for (int fuzz = FUZZ_START; fuzz <= FUZZ_STOP; fuzz += 1) {
        const int len = base_len + fuzz;
        if (len <= 0) continue;
        if (len >= base_len * 2 + fuzz) continue;
        if (! test (generator, len)) {
          printf ("\r%s FAILED with length %08d\n", name, len);
          fflush (stdout);
          return;
        }
        else {
          printf ("\r%s: length %08d OK", name, len);
          fflush (stdout);
        }
      }

    printf ("\r%s: SUCCESS                                \n", name);
  }


  void run_matrix_test(float_gen &generator,
                       const char *name,
                       bool test(float_gen &, const unsigned, const unsigned))
  {
    for (int base_rows = 1; base_rows <= MAX_MATRIX_ROWS; base_rows *= 2) {
      for (int row_fuzz = FUZZ_START; row_fuzz <= FUZZ_STOP; row_fuzz += 1) {
        const int rows = base_rows + row_fuzz;
        if (rows <= 0) continue;
        if (rows >= base_rows * 2 + row_fuzz) continue;
        
        for (int base_cols = 1; base_cols <= MAX_MATRIX_COLS; base_cols *= 2) {
          for (int col_fuzz = FUZZ_START; col_fuzz <= FUZZ_STOP; col_fuzz += 1) {
            const int cols = base_cols + col_fuzz;
            if (cols <= 0) continue;
            if (cols >= base_cols * 2 + col_fuzz) continue;
            if (rows * cols > MAX_NUM_ELEMS) continue;

            if (! test (generator, rows, cols)) {
              printf ("\r%s FAILED with dimensions (%08d, %08d)\n",
                      name,
                      rows,
                      cols);
              fflush (stdout);
              return;
            }
            else {
              printf ("\r%s: dimensions (%08d, %08d) OK",
                      name, rows, cols);
              fflush (stdout);
            }
          }
        }
      }
    }

    printf ("\r%s: SUCCESS                                \n", name);
  }
} /* end anonymous namespace */



//--------------------------------------------------------------------
// MAIN
//--------------------------------------------------------------------
int main(int argc, char * argv[])
{
  float_gen fgen (1, 2);

  printf("Array tests:\n");
  run_array_test (fgen, "forward difference", test_forward_diff);
  run_array_test (fgen, "SAXPY", test_saxpy);
  run_array_test (fgen, "add 1", test_add1);
  run_array_test (fgen, "RMSE", test_rmse);
  // run_array_test (fgen, "add sum", test_add_sum);
  run_array_test (fgen, "Black-Scholes", test_black_scholes);

  printf("Matrix tests:\n");
  run_matrix_test (fgen, "add 1", test_add1_2d);
  run_matrix_test (fgen, "add matrices", test_add_matrices);
  run_matrix_test (fgen, "add 3 matrices", test_add_matrices3);
  run_matrix_test (fgen, "row shift multiply", test_row_shift_multiply);
  run_matrix_test (fgen, "col shift multiply", test_col_shift_multiply);
  run_matrix_test (fgen, "jacobi stencil", test_jacobi_stencil);
  run_matrix_test (fgen, "matrix sum", test_matrix_sum);
  run_matrix_test (fgen, "matrix max", test_matrix_max);
  run_matrix_test (fgen, "jacobi diff", test_jacobi_diff);

  return 0;
}
